module std.thread;


class ThreadError : Exception
{
    this( char[] msg )
    {
        super( msg );
    }
}


private
{
    template min(T)
    {
        T min( T lhs, T rhs ) { return rhs < lhs ? rhs : lhs; }
    }

    template max(T)
    {
        T max( T lhs, T rhs ) { return lhs > rhs ? lhs : rhs; }
    }

    version( X86 )
    {
        enum
        {
            edi,
            esi,
            ebp,
            esp,
            ebx,
            edx,
            ecx,
            eax
        }
    }
}


version( Win32 )
{
    private
    {
        import sys.c.windows.process;
        import sys.c.windows.windows;
        import std.c.string; // for memset

        //
        // entry point for Win32 threads
        //
        extern (Windows) uint threadFunc( void* arg )
        {
	        Thread  obj = cast(Thread) arg;
	        // maybe put an auto exception object here (using alloca) for OutOfMemoryError
	        // plus something to track whether an exception is in-flight?

            obj.m_bstack = getStackBottom();
	        obj.m_tstack = obj.m_bstack;
	        Thread.add( obj );
	        try
	        {
	            obj.run();
	        }
	        catch( Object e )
	        {
	            // error should really print to stderr
	        }
	        Thread.remove( obj );
	        return 0;
        }

        //
        // copy of the same-named function in phobos.std.thread,
        // it uses the Windows naming convention to be consistent
        // with GetCurrentThreadId
        //
        HANDLE GetCurrentThreadHandle()
        {
            const uint DUPLICATE_SAME_ACCESS = 0x00000002;

    	    HANDLE curr = GetCurrentThread(),
    	           proc = GetCurrentProcess(),
    	           hndl;

    	    DuplicateHandle( proc, curr, proc, &hndl, 0, TRUE, DUPLICATE_SAME_ACCESS );
    	    return hndl;
        }

        void* getStackBottom()
        {
            asm
            {
        	    naked;
        	    mov	EAX, FS:4;
        	    ret;
            }
        }

        void* getStackTop()
        {
        	asm
        	{   naked;
        	    mov EAX, ESP;
        	    ret;
        	}
        }

        const uint INFINITE      = 0xFFFFFFFF;
        const uint WAIT_OBJECT_0 = 0x00000000;
        const uint STILL_ACTIVE  = 0x00000103;
    }

    //
    // Thread class
    //
    class Thread
    {
        this()
        {
            m_call = Call.NO;
        }

        this( void function() fn )
        {
            m_fn   = fn;
            m_call = Call.FN;
        }

        this( void delegate() dg )
        {
            m_dg   = dg;
            m_call = Call.DG;
        }

        ~this()
        in
        {
            assert( !isRunning() );
        }
        body
        {
            if( m_addr )
                remove( this );
            CloseHandle( m_hndl );
        }

        final char[] name()
        {
            synchronized( this )
            {
                return m_name;
            }
        }

        final void name( char[] n )
        {
            synchronized( this )
            {
                m_name = n.dup;
            }
        }

        final void start()
        in
        {
            assert( !m_hndl && !m_addr );
        }
        body
        {
            m_hndl = cast(HANDLE) _beginthreadex( null, 0, &threadFunc, this, 0, &m_addr );
            if( cast(size_t) m_hndl == 0 )
                throw new ThreadError( "Error creating thread" );
        }

        final bit isRunning()
        {
            if( !m_addr )
                return false;
	        uint ecode = 0;
	        GetExitCodeThread( m_hndl, &ecode );
	        return ecode == STILL_ACTIVE;
        }

        final void join()
        {
            if( WaitForSingleObject( m_hndl, INFINITE ) != WAIT_OBJECT_0 )
                throw new ThreadError( "Unable to join thread" );
        }

        final void suspend()
        {
            // NOTE: this operation uses the global shared lock for the following reasons:
            //         - typically, this function is called by suspendAll during a GC run,
            //           and it would be prohibitively expensive to have a separate synch
            //           block for every thread that needs to be suspended
            //         - implementing a correct lockless per-thread synchronization mechanism
            //           is too complicated to be worthwhile
            //         - this prevents the unlikely chance that a thread may try to suspend
            //           another thread that is beginning a GC run, thus deadlocking the app
            //       this implementation also relies on synch blocks being reentrant, as it
            //       is quite likely that a GC run will first lock the thread list and then
            //       attempt to pause each thread
            //
            // NOTE: it might be useful to use DBC to ensure that the suspending thread is
            //       the only thread that accesses this thread's stack pointers and such.
            synchronized( slock )
            {
                if( m_scount > 0 )
                {
                    ++m_scount;
                    return;
                }
                // NOTE: atempting to suspend a non-running thread will throw an exception.
                //       is this reasonable?
                if( m_addr != GetCurrentThreadId() && SuspendThread( m_hndl ) == 0xFFFFFFFF )
                    throw new ThreadError( "Unable to suspend thread" );
                ++m_scount;

    	        CONTEXT context;
    	        context.ContextFlags = CONTEXT_INTEGER | CONTEXT_CONTROL;

    	        if( !GetThreadContext( m_hndl, &context ) )
    	            throw new ThreadError( "Unable to load thread context" );
                m_tstack = cast(void*) context.Esp;
                // edi,esi,ebp,esp,ebx,edx,ecx,eax
                m_reg[0] = context.Edi;
                m_reg[1] = context.Esi;
                m_reg[2] = context.Ebp;
                m_reg[3] = context.Esp;
                m_reg[4] = context.Ebx;
                m_reg[5] = context.Edx;
                m_reg[6] = context.Ecx;
                m_reg[7] = context.Eax;
                /+
                memcpy( &m_reg[0], &context.Edi, uint.sizeof * 2 );
                m_reg[2] = context.Ebp;
                m_reg[3] = context.Esp;
                memcpy( &m_reg[4], &context.Ebx, uint.sizeof * 4 );
                +/
            }
        }

        final void resume()
        {
            // NOTE: see comments in suspend for why this shared lock is used
            synchronized( slock )
            {
                if( m_scount < 1 )
                    return;
                if( m_scount > 1 )
                {
                    --m_scount;
                    return;
                }
                // NOTE: atempting to resume a non-running thread will throw an exception.
                //       is this reasonable?
                if( m_addr != GetCurrentThreadId() && ResumeThread( m_hndl ) == 0xFFFFFFFF )
                    throw new ThreadError( "Unable to resume thread" );
                --m_scount;

                m_tstack = m_bstack;
                memset( &m_reg[0], 0, uint.sizeof * m_reg.length );
            }
        }

        final void* stackTop()
        {
            return m_tstack;
        }

        final void* stackBottom()
        {
            return m_bstack;
        }

        final void* regFirst()
        {
            return &m_reg[0];
        }

        final void* regLast()
        {
            return &m_reg[length-1];
        }

        static this()
        {
            Thread main   = new Thread();
            main.m_addr   = GetCurrentThreadId();
            main.m_hndl   = GetCurrentThreadHandle();
            main.m_bstack = getStackBottom();
            main.m_tstack = main.m_bstack;
            add( main );
        }

        static ~this()
        {
            foreach( Thread t; Thread )
            {
                if( !t.isRunning() )
                    remove( t );
            }
        }

        static void sleep( uint milliseconds )
        {
            Sleep( milliseconds );
        }

        static void yield()
        {
            Sleep( 0 );
        }

        static size_t count()
        {
            synchronized( slock )
            {
                return sm_all.length;
            }
        }

        static void suspendAll()
        {
            foreach( Thread curr; Thread )
            {
                if( curr.isRunning() )
                    curr.suspend();
                else
                    remove( curr );
            }
        }

        static void resumeAll()
        {
            foreach( Thread curr; Thread )
            {
                if( curr.isRunning() )
                    curr.resume();
                else
                    remove( curr );
            }
        }

        static Thread getThis()
        {
            synchronized( slock )
            {
                return sm_all[GetCurrentThreadId()];
            }
        }

        static Thread[] getAll()
        {
            synchronized( slock )
            {
                // NOTE: returning sm_all.values involves DMA
                //       which is unsafe if called from gc code
                return sm_all.values;
            }
        }

        static int opApply( int delegate( inout Thread ) dg )
        {
            synchronized( slock )
            {
                int ret = 0;

                foreach( Thread curr; sm_all )
                {
                    ret = dg( curr );
                    if( ret )
                        break;
                }
                return ret;
            }
        }

    protected:
        void run()
        {
            switch( m_call )
            {
            case Call.FN:
                m_fn();
                break;
            case Call.DG:
                m_dg();
                break;
            default:
                break;
            }
        }

    private:
        static Object slock()
        {
            return Thread.classinfo;
        }

        static void add( Thread t )
        in
        {
            assert( t.m_addr );
        }
        body
        {
            synchronized( slock )
            {
                sm_all[t.m_addr] = t;
            }
        }

        static void remove( Thread t )
        in
        {
            assert( t.m_addr );
        }
        body
        {
            synchronized( slock )
            {
                sm_all.remove( t.m_addr );
            }
        }

        enum Call
        {
            NO,
            FN,
            DG
        }

        // indexed by addr
        static Thread[uint] sm_all;

        HANDLE              m_hndl;
        uint                m_addr;
        void*               m_bstack,
                            m_tstack;
        uint[8]             m_reg; // edi,esi,ebp,esp,ebx,edx,ecx,eax
        int                 m_scount;
        Call                m_call;
        char[]              m_name;
        union
        {
            void function() m_fn;
            void delegate() m_dg;
        }
    }

    extern (C) size_t threadCount()
    {
        return Thread.count;
    }

    extern (C) void suspendAllThreads()
    {
        Thread.suspendAll();
    }

    extern (C) void resumeAllThreads()
    {
        Thread.resumeAll();
    }

    private alias void delegate( void*, void* ) scanAllThreadsFn;

    extern (C) void scanAllThreads( scanAllThreadsFn fn )
    {
        foreach( Thread t; Thread )
        {
            // pointers should always refer to a valid range
            assert( t.stackTop && t.stackBottom );
            // stack grows down on x86 machines
            assert( t.stackTop <= t.stackBottom );
            fn( t.stackTop, t.stackBottom + 1 );

            // pointers should always refer to a valid range
            assert( t.regFirst && t.regLast );
            // regFirst should always be the min address
            assert( t.regFirst <= t.regLast );
	        fn( t.regFirst, t.regLast + 1 );
        }
    }
}
else
{
    private
    {
        import std.c.posix.semaphore;
        import std.c.posix.pthread;
        import std.c.posix.signal;
        import std.c.posix.time;
        import sys.c.linux.linux;

        //
        // entry point for POSIX threads
        //
        extern (C) void* threadFunc( void* arg )
        {
	        Thread  obj = cast(Thread) arg;
	        // maybe put an auto exception object here (using alloca) for OutOfMemoryError
	        // plus something to track whether an exception is in-flight?

            obj.m_bstack = getStackBottom();
	        obj.m_tstack = obj.m_bstack;
	        Thread.add( obj );
	        try
	        {
	            obj.run();
	        }
	        catch( Object e )
	        {
	            // error should really print to stderr
	        }
	        Thread.remove( obj );
	        return null;
        }

        extern (C) void suspendHandler( int sig )
        in
        {
            assert( sig == SIGUSR1 );
        }
        body
        {
            Thread* obj = Thread.getThis();
            assert( obj );

            obj.m_tstack = getStackTop();
            asm
            {
                pushad;
            }
            assert( getStackTop() < obj.m_tstack );
            obj.m_freg = obj.m_tstack - 1;
            obj.m_lreg = getStackTop();

            // signal suspended
            // wait for wakeup signal

            obj.m_lreg = null;
            obj.m_freg = null;
	        asm
	        {
	            popad;
	        }
	        obj.m_tstack = obj.m_bstack;
        }
        /+
        extern (C) void resumeHandler( int sig )
        in
        {
            assert( sig == SIGUSR2 );
        }
        body
        {

        }
        +/
        //
        // NOTE: this may not work on all versions of linux,
        //       but apparently the Windows implementation
        //       does not work in Linux (according to DMD revs)
        //
        void* getStackBottom()
        {
            return __libc_stack_end;
        }

        void* getStackTop()
        {
        	asm
        	{   naked;
        	    mov EAX, ESP;
        	    ret;
        	}
        }
    }

    //
    // Thread class
    //
    class Thread
    {
        this()
        {
            m_call      = Call.NO;
        }

        this( void function() fn )
        {
            m_fn        = fn;
            m_call      = Call.FN;
        }

        this( void delegate() dg )
        {
            m_dg        = dg;
            m_call      = Call.DG;
        }

        ~this()
        in
        {
            assert( !isRunning() );
        }
        body
        {
            if( m_addr )
                remove( this );
            pthread_detach( m_addr );
        }

        final char[] name()
        {
            synchronized( this )
            {
                return m_name;
            }
        }

        final void name( char[] n )
        {
            synchronized( this )
            {
                m_name = n.dup;
            }
        }

        final void start()
        in
        {
            assert( !m_hndl && !m_addr );
        }
        body
        {
            if( pthread_create( &m_addr, null, &threadFunc, this ) != 0 )
                throw new ThreadError( "Error creating thread" );
        }

        final bit isRunning()
        {
            // NOTE: possible implementations include registering a cleanup
            //       handler so we can detect thread cancel and manually
            //       tracking whether the thread function has exited or
            //       looking for a pthread function that can tell us this
            //       (possibly by checking for 'thread id invalid' return val)
        }

        final void join()
        {
            if( pthread_join( m_addr, null ) != 0 )
                throw exception( "Unable to join thread" );
        }

        final void suspend()
        {
            // NOTE: this operation uses the global shared lock for the following reasons:
            //         - typically, this function is called by suspendAll during a GC run,
            //           and it would be prohibitively expensive to have a separate synch
            //           block for every thread that needs to be suspended
            //         - implementing a correct lockless per-thread synchronization mechanism
            //           is too complicated to be worthwhile
            //         - this prevents the unlikely chance that a thread may try to suspend
            //           another thread that is beginning a GC run, thus deadlocking the app
            //       this implementation also relies on synch blocks being reentrant, as it
            //       is quite likely that a GC run will first lock the thread list and then
            //       attempt to pause each thread
            //
            // NOTE: it might be useful to use DBC to ensure that the suspending thread is
            //       the only thread that accesses this thread's stack pointers and such.
            synchronized( slock )
            {
                if( m_scount > 0 )
                {
                    ++m_scount;
                    return;
                }
                if( m_addr == pthread_self() )
                {
                    ++m_scount;

                    m_tstack = getStackTop();
    	            asm
    	            {
    	                pushad;
    	            }
    	            assert( getStackTop() < m_tstack );
    	            m_freg = m_tstack - 1;
    	            m_lreg = getStackTop();
                }
                else
                {
                    // NOTE: atempting to suspend a non-running thread will throw an exception.
                    //       is this reasonable?
                    if( pthread_kill( m_addr, SIGUSR1 ) != 0 )
                        throw new ThreadError( "Unable to suspend thread" );
                    sem_wait( &m_suspend );
                    ++m_scount;
                }
            }
        }

        final void resume()
        {
            // NOTE: see comments in suspend for why this shared lock is used
            synchronized( slock )
            {
                if( m_scount < 1 )
                    return;
                if( m_scount > 1 )
                {
                    --m_scount;
                    return;
                }
                if( m_addr == pthread_self() )
                {
                    // pop registers off the stack before resuming
	                asm
	                {
	                    popad;
	                }
                }
                else
                {
                    // NOTE: atempting to resume a non-running thread will throw an exception.
                    //       is this reasonable?
                    if( pthread_kill( m_addr, SIGUSR2 ) != 0 )
                        throw new ThreadError( "Unable to resume thread" );
                }
                --m_scount;

                m_tstack = m_bstack;
                m_breg   = null;
                m_treg   = null;
            }
        }

        final void* stackTop()
        {
            return m_tstack;
        }

        final void* stackBottom()
        {
            return m_bstack;
        }

        final void* regFirst()
        {
            return m_freg;
        }

        final void* regLast()
        {
            return m_lreg;
        }

        static this()
        {
            int         status;
            sigaction_t sigusr1,
                        sigusr2;

            sigusr1.sa_flags   = 0;
            sigusr1.sa_handler = &suspendHandler;
            sigemptyset( &sigusr1.sa_mask );

            /+
            sigusr2.sa_flags   = 0;
            sigusr2.sa_handler = &resumeHandler;
            sigemptyset( &sigusr2.sa_mask );
            +/
            status = sigaction( SIGUSR1, &sigusr1, null );
            assert( status == 0 );
            /+
            status = sigaction( SIGUSR2, &sigusr2, null );
            assert( status == 0 );
            +/
            status = pthread_key_create( &sm_this, null );
            assert( status == 0 );

            Thread main   = new Thread();
            main.m_addr   = pthread_self();
            main.m_bstack = getStackBottom();
            main.m_tstack = main.m_bstack;

            status = pthread_setspecific( sm_this, main );
            assert( status == 0 );

            add( main );
        }

        static ~this()
        {
            foreach( Thread t; Thread )
            {
                if( !t.isRunning() )
                    remove( t );
            }
        }

        static void sleep( uint milliseconds )
        {
            pthread_mutex_t mutex;
            pthread_cond_t  cond;
            timespec        wait;
            timeval         now;

            gettimeofday( &now, null );
            wait.tv_sec = milliseconds / 1000 + now.tv_sec;
            wait.tv_nsec = milliseconds % 1000 * 1000;
            while( wait.tv_nsec > 1000000000 )
            {
                wait.tv_nsec -= 1000000000;
                ++dt_sec;
            }
            wait.tv_sec += now.tv_sec;

            pthread_mutex_init( &mutex, null );
            pthread_cond_init( &cond, null );

            pthread_mutex_lock( &mutex );
            pthread_cond_timedwait( &cond, &mutex, &wait );
            pthread_mutex_unlock( &mutex );

            pthread_mutex_destroy( &mutex );
            pthread_cond_destroy( &cond );
        }

        static void yield()
        {
            sched_yield();
        }

        static size_t count()
        {
            // NOTE: this could be made a tad more efficient by using
            //       a separate lockless counter, at the cost of
            //       more expensive writes
            synchronized( slock )
            {
                return sm_all.length;
            }
        }

        static void suspendAll()
        {
            foreach( Thread curr; Thread )
            {
                if( curr.isRunning() )
                    curr.suspend();
                else
                    remove( curr );
            }
        }

        static void resumeAll()
        {
            foreach( Thread curr; Thread )
            {
                if( curr.isRunning() )
                    curr.resume();
                else
                    remove( curr );
            }
        }

        static Thread getThis()
        {
            // NOTE: this function may not be called until the module
            //       has been initialized.  this should not be an issue
            //       given the current implementation.  there is no
            //       precondition to test this because I have no idea
            //       whether zero is a legal value for an initialized
            //       pthread_key_t
            return cast(Thread) pthread_getspecific( sm_this );
        }

        static Thread[] getAll()
        {
            synchronized( slock )
            {
                // NOTE: returning sm_all.values involves DMA
                //       which is unsafe if called from gc code
                return sm_all.values;
            }
        }

        static int opApply( int delegate( inout Thread ) dg )
        {
            synchronized( slock )
            {
                int ret = 0;

                foreach( Thread curr; sm_all )
                {
                    ret = dg( curr );
                    if( ret )
                        break;
                }
                return ret;
            }
        }

    protected:
        void run()
        {
            switch( m_call )
            {
            case Call.FN:
                m_fn();
                break;
            case Call.DG:
                m_dg();
                break;
            default:
                break;
            }
        }

    private:
        static Object slock()
        {
            return Thread.classinfo;
        }

        static void add( Thread t )
        in
        {
            assert( t.m_addr );
        }
        body
        {
            synchronized( slock )
            {
                sm_all[t.m_addr] = t;
            }
        }

        static void remove( Thread t )
        in
        {
            assert( t.m_addr );
        }
        body
        {
            synchronized( slock )
            {
                sm_all.remove( t.m_addr );
            }
        }

        enum Call
        {
            NO,
            FN,
            DG
        }

        static pthread_key_t        sm_this;
        static Thread[pthread_t]    sm_all;

        pthread_t   m_addr;
        void*       m_bstack,
                    m_tstack,
                    m_freg,
                    m_lreg;
        int         m_scount;
        Call        m_call;
        char[]      m_name;
        union
        {
            void function() m_fn;
            void delegate() m_dg;
        }
    }

    extern (C) size_t threadCount()
    {
        return Thread.count;
    }

    extern (C) void suspendAllThreads()
    {
        Thread.suspendAll();
    }

    extern (C) void resumeAllThreads()
    {
        Thread.resumeAll();
    }

    private alias void delegate( void*, void* ) scanAllThreadsFn;

    extern (C) void scanAllThreads( scanAllThreadsFn fn )
    {
        foreach( Thread t; Thread )
        {
            // pointers should always refer to a valid range
            assert( t.stackTop && t.stackBottom );
            // stack grows down on x86 machines
            assert( t.stackTop <= t.stackBottom );
            fn( t.stackTop, t.stackBottom + 1 );

            // pointers should always refer to a valid range
            assert( t.regFirst && t.regLast );
            // regFirst should always be the min address
            assert( t.regFirst <= t.regLast );
	        fn( t.regFirst, t.regLast + 1 );
        }
    }
}